package no.ntnu.item.ramses.runtime;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import no.ntnu.item.arctis.runtime.AbstractRamsesRuntime;
import no.ntnu.item.arctis.runtime.IFilterableSession;
import no.ntnu.item.ramses.runtime.logging.Logger;
import no.ntnu.item.ramses.runtime.routing.Router;

public class Scheduler extends AbstractRamsesRuntime implements Runnable {

    private static final boolean debug = true;

    /**
     * The Scheduler will keep looping, looking for transactions, as long as
     * keepRunning is true.
     */
    private boolean keepRunning;

    /**
     * The router this Scheduler forwards and receives its messages to/from.
     */
    private final Router router;

    /**
     * An address book for the state machines running on the Scheduler. It
     * associates both the state machine IDs, addresses and component types with
     * state machines.
     */
    private final LocalStateMachineRegistry localStateMachines;

    /**
     * The component manager keeps track of components who have sessions.
     */
    private final ComponentRegistry componentRegistry;

    /**
     * The messageQueues is a table that contains the message queues of all the
     * state machines. The state machine instance is the key to the table, and
     * its message queue the value.
     */
    private final Hashtable messageQueues;

    /**
     * The Vector timers contains all the active timers in the Scheduler.
     */
    private final Vector timers;

    /**
     * All state machines that are not yet initiated is kept in this Vector.
     */
    private final Vector uninitatedStateMachines;
    
    /**
     * The generator for random session IDs.
     */
    protected final Random randomSessionIdGenerator;

    public Scheduler(Router router) {
        uninitatedStateMachines = new Vector();
        localStateMachines = new LocalStateMachineRegistry();
        componentRegistry = new ComponentRegistry();

        messageQueues = new Hashtable();

        timers = new Vector();
        
        randomSessionIdGenerator = new Random();

        this.router = router;
    }

    public String registerMainStateMachine(String componentType, IStateMachine sm) {
        debug("Registering main state machine of type " + componentType, "registerMainStateMachine");
    
        Object[] smAddresses = router.registerMainStateMachine(componentType, this);
        String smId = router.extractComponentId(smAddresses[0]);
        
        localStateMachines.put(sm, smId, smAddresses[0]);
        
        for (int i = 1; i < smAddresses.length; i++) {
            localStateMachines.addExtraAddress(smAddresses[i], sm);
        }
    
        messageQueues.put(sm, new InputQueue());
        uninitatedStateMachines.addElement(sm);
    
        debug("Statemachine, " + smId + ", registered.", "registerMainStateMachine");
        return smId;
    }

    public String registerSessionStateMachine(IStateMachine outerComponent, String partName, String sessionId,
            IStateMachine sm) {
        debug("Registering session state machine for session ID " + sessionId, "registerSessionsStateMachine");
        debug("Session state machine is of class: " + sm.getClass().toString(), "registerSessionStateMachine");

        Object[] smAddresses = router.registerSessionStateMachine(outerComponent.getId(), partName, sessionId, this);
        String smId = router.extractSessionSmId(smAddresses[0]);
        
        localStateMachines.put(sm, smId, smAddresses[0]);
        
        for (int i = 1; i < smAddresses.length; i++) {
            localStateMachines.addExtraAddress(smAddresses[i], sm);
        }

        messageQueues.put(sm, new InputQueue());
        uninitatedStateMachines.addElement(sm);

        debug("Session state machine registered, session ID: " + sessionId, "registerStateMachine");
        
        return smId;
    }

    public void registerActionClass(Object actionClass, Object stateMachine) {
        if (!(stateMachine instanceof IStateMachine)) {
            throw new IllegalArgumentException("Invalid State machine object.");
        }
    
        instances.put(actionClass, this);
    
        IStateMachine stm = (IStateMachine)stateMachine;
        debug("Registering action: " + actionClass + "on statemachine:" + stm.getId(), "registerActionClass");
    
        localStateMachines.registerActionClass(stm, actionClass);
    }

    public void deregisterStateMachine(IStateMachine sm) {
        debug("Deregistering state machine: " + sm.getId(), "deregisterStateMachine");

        // If state machine contains sessions, all those should be removed.
        Vector allSessions = componentRegistry.getAllSessionsInComponent(sm);
        for (Enumeration e = allSessions.elements(); e.hasMoreElements();) {
            deregisterStateMachine((IStateMachine)e.nextElement());
        }

        router.deregisterStateMachine(localStateMachines.getAddressOfStateMachine(sm));
        if (componentRegistry.getSessionIdOfStateMachine(sm) != null) componentRegistry.removeSession(sm);
        if (componentRegistry.isRegisteredComponent(sm)) componentRegistry.deregisterComponent(sm);
        localStateMachines.remove(sm);
        messageQueues.remove(sm);
        
        Vector timersToRemove = new Vector();
        for (int i = (timers.size() - 1); i >= 0; i--) {
            Timer t = (Timer)timers.get(i);
            if (t.stateMachine == sm) {
                timersToRemove.add(i);
            }
        }
        while (!timersToRemove.isEmpty()) {
            timers.remove(((Integer)timersToRemove.firstElement()).intValue());
            timersToRemove.remove(0);
        }
        
        if (localStateMachines.isEmpty()) {
            debug("No more state machines to execute, closing scheduler!", "deregisterStateMachine");
            destroy();
        }
    }

    public void run() {
        /*
         * This method implements a prioritized round robin scheduling scheme
         * with the following prioritization: (1) Timers, (2) Uninitiated and
         * registered state machines, (3) Normal messages
         */
        keepRunning = true;
        debug("Scheduler starts running.", "run");

        while (keepRunning) {

            /*
             * Check for and execute expired timers.
             */
            long currentTime = System.currentTimeMillis();
            while (!timers.isEmpty() && ((Timer)timers.firstElement()).time <= currentTime) {
                Timer timer = (Timer)timers.firstElement();
                timers.removeElementAt(0);
                timer.stateMachine.fireTimer(timer.id);
            }

            /*
             * Check for and initate uninitiated state machines.
             */
            while (uninitatedStateMachines.size() > 0) {
                debug("Initiating state machine(s).", "run");
                IStateMachine sm = (IStateMachine)uninitatedStateMachines.elementAt(0);
                sm.fireInitial();
                uninitatedStateMachines.removeElementAt(0);
            }

            /*
             * Execute one transition for the state machine that is next in the
             * round robin-scheme.
             */
            IStateMachine nextSm = localStateMachines.nextStateMachine();
            if (nextSm != null) {
                boolean consumed = false;
                boolean discarded = false;
                InputQueue queue = getInputQueue(nextSm);
                Message msg = queue.peakAtFirstNonMarkedMessage();
                while (!consumed && !discarded && (msg != null)) {
                    int result = nextSm.fire(msg.sender, msg.sessionId, msg.signalID, msg.content);
                    if (result == IStateMachine.CONSUME_SIGNAL) {
                        queue.removeMessage(msg);
                        queue.unmarkAllMessages();
                        debug("consumed", "run");
                        consumed = true;
                    }
                    else if (result == IStateMachine.DISCARD_SIGNAL) {
                        queue.removeMessage(msg);
                        discarded = true;
                        Logger.getLogger().discardedSignal(msg.signalID, msg.sender.toString(),
                                msg.receiver.toString(), nextSm.getClass().getName());
                    }
                    else {
                        debug("marked as saved", "run");
                        msg.markAsSaved(true);
                        msg = queue.peakAtFirstNonMarkedMessage();
                    }
                }
            }
            else {
                debug("Warning: Scheduler is running, but could not find a state machine to execute!", "run");
            }

            if (uninitatedStateMachines.isEmpty() && allInputQueuesAreEmpty() && keepRunning) {
                /*
                 * Wait untill next timer expires OR an incoming message
                 * notifies the object, see deliver().
                 */
                debug("No initiated state machines and all input queues are empty, will wait.", "run");
                try {
                    synchronized (this) {
                        if (!timers.isEmpty()) {
                            long wait = ((Timer)timers.firstElement()).time - System.currentTimeMillis();
                            if (wait > 10) {
                                debug("Waits untill next timer expires: " + wait + "ms.", "run");
                                wait(wait);
                            }
                        }
                        else {
                            debug("Waits indefinately, untill external message is delivered.", "run");
                            wait();
                        }
                    }
                }
                catch (InterruptedException e) {
                    // Ignore.
                }
            }
        }
    }

    public void send(Object sender, Object receiver, String signalId, Object data) {
        debug("Entering", "send");
        
        IStateMachine senderSm = localStateMachines.get(sender);
        Object senderAddress = localStateMachines.getAddressOfStateMachine(senderSm);
        
        if (receiver instanceof String) receiver = router.createAddress((String)receiver);
        else if (receiver instanceof IStateMachine) receiver = localStateMachines.getAddressOfStateMachine((IStateMachine)receiver);
        Object componentMainSmAddress = router.extractComponentAddress(receiver);
        String partName = router.extractPartName(receiver);
        String sessionId = router.extractSessionId(receiver);
        
        deliver(senderAddress, componentMainSmAddress, partName, sessionId, false, signalId, data, false);
        
        debug("Leaving", "send");
    }

    public void send(Object sender, Object receiver, String id) {
        send(sender, receiver, id, null);
    }

    public void sendToMe(Object actionClass, String signalId, Object data) {
        debug("Entering", "sendToMe");
    
        IStateMachine sm = (IStateMachine)localStateMachines.get(actionClass);
        if (sm == null) {
            throw new IllegalArgumentException("Action class not registered: " + actionClass);
        }
        
        debug("Found state machine, sending message.", "sendToMe");
        
        Object senderAddress = localStateMachines.getAddressOfStateMachine(sm);
        Object componentMainSmAddress = router.extractComponentAddress(senderAddress);
        String partName = router.extractPartName(senderAddress);
        String sessionId = router.extractSessionId(senderAddress);
    
        deliver(senderAddress, componentMainSmAddress, partName, sessionId, false, signalId, data, false);
    
        debug("Leaving", "sendToMe");
    }

    public void sendToMe(Object actionClass, String signalId) {
        sendToMe(actionClass, signalId, null);
    }

    public void sendToNewLocalSession(Object sender, String partName, String signalId, Object data, Hashtable properties) {
        debug("Entering..", "sendToNewLocalSession");
        
        IStateMachine senderSm = localStateMachines.get(sender);
        Object senderAddress = localStateMachines.getAddressOfStateMachine(senderSm);
        
        String sessionId = (String)getFreshSessionId();
        
        deliver(senderAddress, senderAddress, partName, sessionId, true, signalId, data, false);
        
        IStateMachine sessionSm = componentRegistry.getSession(senderSm, partName, sessionId);
        if (properties != null && sessionSm != null) {
            debug("Setting properties", "sendToNewLocalSession");

            for (Enumeration e = properties.keys(); e.hasMoreElements();) {
                String id = (String)e.nextElement();
                sessionSm.setProperty(id, properties.get(id));
            }
        }
        
        debug("Leaving..", "sendToNewLocalSession");
    }

    public void sendToNewLocalSession(Object sender, String partName, String signalId, Hashtable properties) {
        sendToNewLocalSession(sender, partName, signalId, null, properties);
    }

    public Object sendToNewRemoteSession(Object sender, Object outerComponent, String partName, Object sessionId, String signalId, Object data) {
        debug("Entering..", "sendToNewRemoteSession");
        
        IStateMachine senderSm = localStateMachines.get(sender);
        Object senderAddress = localStateMachines.getAddressOfStateMachine(senderSm);
        
        if (outerComponent instanceof String) outerComponent = router.createAddress((String)outerComponent);
        
        deliver(senderAddress, outerComponent, partName, (String)sessionId, true, signalId, data, false);
        
        debug("Leaving..", "sendToNewRemoteSession");
        return router.getCompleteUrl(outerComponent, partName, (String)sessionId, false);
    }
    
    public Object sendToNewRemoteSession(Object sender, Object outerComponent, String partName, Object sessionId, String signalId) {
        return sendToNewRemoteSession(sender, outerComponent, partName, sessionId, signalId, null);
    }
    
    public void sendToComponentMainStateMachine(Object sender, String signalId, Object data) {
        debug("Entering..", "sendToComponentMainStateMachine");
        
        IStateMachine senderSm = localStateMachines.get(sender);
        Object senderAddress = localStateMachines.getAddressOfStateMachine(senderSm);
        
        IStateMachine componentMainSm = componentRegistry.getOuterComponent(senderSm);
        Object componentMainSmAddress = localStateMachines.getAddressOfStateMachine(componentMainSm);
        
        deliver(senderAddress, componentMainSmAddress, null, null, false, signalId, data, false);
       
        debug("Leaving..", "sendToComponentMainStateMachine");
    }
    
    public void sendToComponentMainStateMachine(Object sender, String signalId) {
        sendToComponentMainStateMachine(sender, signalId, null);
    }

    public void deliver(Object senderAddr, Object componentMainSmAddress, String partName, String sessionId,
            boolean allowCreateSession, String signalId, Object data, boolean fromRouter) {
        debug("Entering", "deliver");

        IStateMachine componentMainSm = localStateMachines.get(componentMainSmAddress);

        if (componentMainSm == null) {
            if (fromRouter) {
                Logger.getLogger().messageDroppedDuringRouting(senderAddr.toString(), componentMainSmAddress.toString(), signalId);
                debug("Got message to non-existant component from router, discarding.", "deliver");
            }
            else {
                router.routeMessage(senderAddr, componentMainSmAddress, partName, sessionId, allowCreateSession,
                        signalId, data);
            }
        }
        else {
            if (partName == null) {
                putMessageInInputQueue(senderAddr, componentMainSmAddress, sessionId, signalId, data);
            }
            else {
                if (sessionId == null) {
                    Logger.getLogger().messageDroppedDuringRouting(senderAddr.toString(), componentMainSm.getId() + "/" + partName + "?sessionId=" + null, signalId);
                }
                else {
                    IStateMachine sessionSm = componentRegistry.getSession(componentMainSm, partName, sessionId);
                    if (sessionSm != null) {
                        Object sessionAddr = localStateMachines.getAddressOfStateMachine(sessionSm);
                        putMessageInInputQueue(senderAddr, sessionAddr, sessionId, signalId, data);
                    }
                    else {
                        if (!allowCreateSession) {
                            Logger.getLogger().messageDroppedDuringRouting(senderAddr.toString(), componentMainSm.getId() + "/" + partName + "?sessionId=" + sessionId, signalId);
                        }
                        else {
                            Object sessionAddr = createSession(componentMainSm, partName, sessionId);
                            putMessageInInputQueue(senderAddr, sessionAddr, sessionId, signalId, data);
                        }
                    }
                }
            }
        }

        debug("Leaving", "deliver");
    }
    
    private void putMessageInInputQueue(Object senderAddress, Object receiverAddress, Object sessionId, String signalId, Object data) {
        Message msg = new Message(senderAddress, receiverAddress,
                sessionId, signalId, data);
    
        IStateMachine senderSm = localStateMachines.get(senderAddress);
        IStateMachine receiverSm = localStateMachines.get(receiverAddress);
        
        if (senderSm != null && receiverSm != null && isSameComponent(senderSm, receiverSm)) {
            debug("Internal message, high priority.", "deliver");
            getInputQueue(receiverSm).addInternalMessage(msg);
        }
        else {
            debug("Normal priority message.", "deliver");
            getInputQueue(receiverSm).addExternalMessage(msg);
        }
        awakenScheduler();
    }

    public void startTimer(int delay, String timerId, Object stateMachine) {
        Timer timer = new Timer(delay, timerId, (IStateMachine)stateMachine);
        insertTimerInSortedVector(timer);
        debug("Starting timer with ID: " + timerId, "startTimer");
    }

    public void stopTimer(String timerId, Object stateMachine) {
        for (int i = 0; i < timers.size(); i++) {
            Timer timer = (Timer)timers.elementAt(i);
            if (timer.id.equals(timerId) && (timer.stateMachine == stateMachine)) {
                timers.removeElementAt(i);
            }
        }
    }

    public void restartTimer(int delay, String timerId, Object stm) {
        for (int i = 0; i < timers.size(); i++) {
            Timer timer = (Timer)timers.elementAt(i);
            if (timer.id.equals(timerId) && (timer.stateMachine == stm)) {
                timers.removeElementAt(i);
                timer.reset();
                insertTimerInSortedVector(timer);
                return;
            }
        }
        startTimer(delay, timerId, stm);
    }

    /**
     * @deprecated
     */
    public void resetTimer(String timerID, Object stm) {
        for (int i = 0; i < timers.size(); i++) {
            Timer timer = (Timer)timers.elementAt(i);
            if (timer.id.equals(timerID) && (timer.stateMachine == stm)) {
                timers.removeElementAt(i);
                timer.reset();
                insertTimerInSortedVector(timer);
            }
        }
    }

    public Object createAddress(String url) {
        return router.createAddress(url);
    }

    public String getMyAddress(Object actionClass) {
        IStateMachine sm = (IStateMachine)localStateMachines.get(actionClass);
        if (sm == null) {
            throw new IllegalArgumentException("Action class not registered: " + actionClass);
        }
        return localStateMachines.getAddressOfStateMachine(sm).toString();
    }
    
    public Object getAddressWithSessionId(Object address, Object sessionId) {
        Object cleanAddress = router.extractComponentAddress(address);
        return router.getCompleteUrl(cleanAddress, null, (String)sessionId, false);
    }

    public Object getFreshSessionId() {
        return "" + Math.abs(randomSessionIdGenerator.nextLong());
    }

    public Object getMySessionId(Object sessionStateMachine) {
        IStateMachine sessionSm = localStateMachines.get(sessionStateMachine);
        return componentRegistry.getSessionIdOfStateMachine(sessionSm);
    }

    public IFilterableSession[] getSessions(Object mainStateMachine, String partName) {
        IStateMachine mainSm = localStateMachines.get(mainStateMachine);
        Vector sessions = componentRegistry.getSessions(mainSm, partName);
        
        if (sessions == null) return new IFilterableSession[0];
        
        IFilterableSession[] filterableSessions = new IFilterableSession[sessions.size()];
        for (int i = 0; i < sessions.size(); i++) {
            IFilterableSession filterableSession = (IFilterableSession)sessions.get(i);
            filterableSessions[i] = filterableSession;
        }
        return filterableSessions;
    }

    public Object getActionClass(IFilterableSession session, String actionClassFqn) {
        IStateMachine sessionSm = (IStateMachine)session;
        return localStateMachines.getActionClassOfType(sessionSm, actionClassFqn);
    }

    public void destroy() {
        keepRunning = false;
        router.notifySchedulerStopped(this);
    }

    private void awakenScheduler() {
        synchronized (this) {
            notifyAll();
        }
    }

    private InputQueue getInputQueue(IStateMachine stm) {
        return (InputQueue)messageQueues.get(stm);
    }

    private boolean allInputQueuesAreEmpty() {
        for (Enumeration e = messageQueues.keys(); e.hasMoreElements();) {
            Object o = e.nextElement();
            if (!((InputQueue)messageQueues.get(o)).isEmpty()) return false;
        }
        return true;
    }

    private void insertTimerInSortedVector(Timer timer) {
        if (timers.size() == 0) {
            timers.addElement(timer);
        }
        else {
            int i = 0;
            boolean endReached = false;
            while (((Timer)timers.elementAt(i)).time < timer.time) {
                // TODO Could make more efficient using binary search or
                // something...
                if (i + 1 == timers.size()) {
                    endReached = true;
                    break;
                }
                else {
                    i++;
                }
            }
            if (!endReached) {
                timers.insertElementAt(timer, i);
            }
            else {
                timers.addElement(timer);
            }
        }
    }

    private Object createSession(IStateMachine outerComponent, String partName, String sessionId) {
        IStateMachine newSession = outerComponent.createSession(partName, sessionId);
    
        if (newSession == null) {
            Logger.getLogger().sessionCreationFailed(partName, outerComponent.getClass().getName());
        }
    
        if (!componentRegistry.isRegisteredComponent(outerComponent)) {
            componentRegistry.registerComponent(outerComponent);
        }
        componentRegistry.registerSession(outerComponent, partName, newSession, sessionId);
    
        return localStateMachines.getAddressOfStateMachine(newSession);
    }

    private boolean isSameComponent(IStateMachine localReceiver, IStateMachine sender) {
        if (localReceiver == sender) return true;

        if (componentRegistry.isRegisteredComponent(localReceiver)) {
            if (componentRegistry.getOuterComponent(sender) == localReceiver) {
                return true;
            }
        }
        else if (componentRegistry.isRegisteredComponent(sender)) {
            if (componentRegistry.getOuterComponent(localReceiver) == sender) {
                return true;
            }
        }
        else if (componentRegistry.getOuterComponent(localReceiver) == componentRegistry.getOuterComponent(sender)) {
            return true;
        }

        return false;
    }

    protected void debug(String text, String method) {
        if (debug) System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
    }

    private static class InputQueue {

        private final Vector externalMessages;
        private final Vector internalMessages;

        public InputQueue() {
            externalMessages = new Vector();
            internalMessages = new Vector();
        }

        public boolean isEmpty() {
            return (externalMessages.isEmpty() && internalMessages.isEmpty());
        }

        public synchronized void addExternalMessage(Message msg) {
            externalMessages.addElement(msg);
        }

        public void addInternalMessage(Message msg) {
            internalMessages.addElement(msg);
        }

        /**
         * @return the first non-marked message or null if there are no such
         *         messages
         */
        public Message peakAtFirstNonMarkedMessage() {
            for (Enumeration iMsgs = internalMessages.elements(); iMsgs.hasMoreElements();) {
                Message nextMsg = (Message)iMsgs.nextElement();
                if (!nextMsg.isMarked()) {
                    return nextMsg;
                }
            }
            for (Enumeration iMsgs = externalMessages.elements(); iMsgs.hasMoreElements();) {
                Message nextMsg = (Message)iMsgs.nextElement();
                if (!nextMsg.isMarked()) {
                    return nextMsg;
                }
            }
            return null;
        }

        public void unmarkAllMessages() {
            for (Enumeration iMsgs = internalMessages.elements(); iMsgs.hasMoreElements();) {
                ((Message)iMsgs.nextElement()).markAsSaved(false);
            }
            for (Enumeration iMsgs = externalMessages.elements(); iMsgs.hasMoreElements();) {
                ((Message)iMsgs.nextElement()).markAsSaved(false);
            }
        }

        public void removeMessage(Message msg) {
            if (internalMessages.contains(msg)) {
                internalMessages.removeElement(msg);
            }
            else if (externalMessages.contains(msg)) {
                externalMessages.removeElement(msg);
            }
            else {
                debug("Warning: Tried to remove message from an InputQueue that did not contain it!", "removeMessage");
            }
        }

        protected void debug(String text, String method) {
            if (debug) System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
        }
    }

    private static class Timer {

        String id;
        int delay;
        long time;
        IStateMachine stateMachine;

        private Timer(int delay, String id, IStateMachine stateMachine) {
            this.delay = delay;
            this.time = System.currentTimeMillis() + delay;
            this.id = id;
            this.stateMachine = stateMachine;
        }

        private void reset() {
            this.time = System.currentTimeMillis() + delay;
        }

    }

    public interface IStateMachine {

        public static final int SAVE_SIGNAL = 0;
        public static final int CONSUME_SIGNAL = 1;
        public static final int DISCARD_SIGNAL = 2;

        public String getId();

        public void fireInitial();

        public int fire(Object sender, Object sessionId, String signalId, Object data);

        public int fireTimer(String timerId);

        public IStateMachine createSession(String partName, String sessionId);

        public void setProperty(String id, Object property);

    }

    /***
     * An address book used to keep track of state machines running on the
     * Scheduler. It keeps track of the state machine IDs, their address (as an
     * Object), their type and any associated action class.
     * 
     * @author Marius Bjerke
     */
    private static class LocalStateMachineRegistry {

        /**
         * The byId table is used to keep track of all the state machines
         * running on the Scheduler by their IDs. The state machine ID is the
         * key to the table, and the value is the corresponding state machine.
         */
        private final Hashtable byId;

        /**
         * The byAddress table is used to keep track of all the state machines
         * running on the Scheduler by their addresses. The state machine
         * address object is the key to the table, and the value is the
         * corresponding state machine.
         */
        private final Hashtable byAddress;

        /**
         * The byActionClass table maps action classes to state machines. It is
         * used to map a signal from an action back to the state machine. The
         * action class is the key to the table, and the corresponding state
         * machine is the value.
         */
        private final Hashtable byActionClass;

        /**
         * The allEntries table keeps records of the information stored about a
         * state machine.
         */
        private final Hashtable allEntries;

        /**
         * The stateMachines vector contains all the state machines, ordered by
         * "next in line for transition execution". This is needed because there
         * is no way to keep them ordered in a Hashtable.
         */
        private final Vector stateMachineQueue;

        /**
         * Initiates an empty address book.
         */
        public LocalStateMachineRegistry() {
            byId = new Hashtable();
            byAddress = new Hashtable();
            byActionClass = new Hashtable();
            allEntries = new Hashtable();
            stateMachineQueue = new Vector();
        }

        public boolean isEmpty() {
            return allEntries.isEmpty();
        }

        /**
         * Stores a state machine in the address book, and associates it with
         * its ID and address.
         * 
         * @param sm
         *            The state machine instance. Must not be null.
         * @param smId
         *            The state machine ID. Must not be null.
         * @param smAddr
         *            The state machine address object. Must not be null.
         */
        public void put(IStateMachine sm, String smId, Object smAddr) {
            if (sm == null || smId == null || smAddr == null)
                throw new IllegalArgumentException("Parameters can not be null.");

            byId.put(smId, sm);

            byAddress.put(smAddr, sm);

            allEntries.put(sm, new AddressBookEntry(smId, smAddr));
            stateMachineQueue.addElement(sm);
        }
        
        public void addExtraAddress(Object smAddr, IStateMachine sm) {
            byAddress.put(smAddr, sm);
        }

        public void remove(IStateMachine sm) {
            if (sm == null) return;
            stateMachineQueue.removeElement(sm);
            byAddress.remove(getAddressOfStateMachine(sm));
            byId.remove(sm.getId());
            allEntries.remove(sm);
        }

        /**
         * Registers an action class to a state machine.
         * 
         * @param sm
         *            The state machine instance, must not be null and should
         *            already be registered in the address book.
         * @param actionClass
         *            The action class to register.
         */
        public void registerActionClass(IStateMachine sm, Object actionClass) {
            if (sm == null || get(sm) == null) {
                throw new IllegalStateException("State machine is null or not registered in address book");
            }

            ((AddressBookEntry)allEntries.get(sm)).actionClasses.addElement(actionClass);
            byActionClass.put(actionClass, sm);
        }

        /**
         * Retrieves a state machine from the LocalAddressBook, or null if the
         * provided key is not a state machine, registered state machine ID or
         * address.
         */
        public IStateMachine get(Object key) {
            if (key instanceof IStateMachine) return (IStateMachine)key;

            Object smObject;
            smObject = byId.get(key);
            if (smObject == null) smObject = byAddress.get(key);
            // if (smObject == null) smObject = byComponentType.get(key);
            if (smObject == null) smObject = byActionClass.get(key);

            if (smObject instanceof IStateMachine)
                return (IStateMachine)smObject;
            else if (smObject == null)
                return null;
            else throw new IllegalStateException("A non-IStateMachine object was once put in the LocalAddressBook!");
        }

        /**
         * Returns the state machines in a round robin fashion.
         * 
         * @return The next IStateMachine according to a round robin scheme.
         */
        public IStateMachine nextStateMachine() {
            if (stateMachineQueue.size() > 0) {
                IStateMachine sm = (IStateMachine)stateMachineQueue.firstElement();
                stateMachineQueue.removeElement(sm);
                stateMachineQueue.addElement(sm);
                return sm;
            }
            else return null;
        }

        /**
         * Gets the registered address of the given state machine.
         * 
         * @param sm
         *            The state machine to find the address for.
         * @return The address of the given state machine, or null if it is not
         *         registered.
         */
        public Object getAddressOfStateMachine(IStateMachine sm) {
            AddressBookEntry entry = (AddressBookEntry)allEntries.get(sm);
            if (entry != null)
                return entry.address;
            else return null;
        }

        /**
         * Gets an action class of a specific type for a state machine.
         * 
         * @param sm
         *            The state machine that has registered the desired action
         *            class object.
         * @param actionClassFqn
         *            The fully qualified name of the desired action class
         *            object.
         * @return The registered action class object of a type with FQN equal
         *         to actionClassFqn
         */
        public Object getActionClassOfType(IStateMachine sm, String actionClassFqn) {
            AddressBookEntry entry = (AddressBookEntry)allEntries.get(sm);
            if (entry == null) return null;

            // TODO: Improve performance with a map?
            Object ac;
            for (Enumeration e = entry.actionClasses.elements(); e.hasMoreElements();) {
                ac = e.nextElement();
                if (ac.getClass().getName().equals(actionClassFqn)) return ac;
            }
            return null;
        }

        private static class AddressBookEntry {

            final String id;
            final Object address;
            final Vector actionClasses;

            public AddressBookEntry(String smId, Object smAddr) {
                id = smId;
                address = smAddr;
                actionClasses = new Vector();
            }
        }
    }

    /**
     * A class to manage and keep track of components, their sessions and other
     * related data.
     * 
     * @author Marius Bjerke
     */
    private static class ComponentRegistry {

        /*
         * Hashtable<String stateMachineId, Component>
         */
        private final Hashtable components;

        /*
         * Hashtable<String sessionStateMachineId, String outerComponentId>
         */
        private final Hashtable outerComponents;

        /*
         * Hashtable<String sessionStateMachineId, String partName>
         */
        private final Hashtable partNames;

        /*
         * Hashtable<String sessionStateMachineId, String sessionId>
         */
        private final Hashtable sessionIds;

        /*
         * Hashtable<String sessionId, IStateMachine session>
         */
        private final Hashtable sessions;

        public ComponentRegistry() {
            components = new Hashtable();
            outerComponents = new Hashtable();
            partNames = new Hashtable();
            sessionIds = new Hashtable();
            sessions = new Hashtable();
        }

        public Vector getAllSessionsInComponent(IStateMachine outerComponent) {
            Vector v = new Vector();
            ComponentEntry c = getComponent(outerComponent.getId());
            if (c != null) {
                for (Enumeration e = c.innerParts.elements(); e.hasMoreElements();) {
                    ComponentEntry.PartEntry p = (ComponentEntry.PartEntry)e.nextElement();
                    for (Enumeration e2 = p.sessions.elements(); e2.hasMoreElements();) {
                        v.addElement(e2.nextElement());
                    }
                }
            }
            return v;
        }

        public void registerComponent(IStateMachine componentMainBehaviour) {
            components.put(componentMainBehaviour.getId(), new ComponentEntry(componentMainBehaviour));
        }

        public boolean isRegisteredComponent(IStateMachine componentMainBehaviour) {
            if (components.containsKey(componentMainBehaviour.getId())) return true;
            return false;
        }

        public void deregisterComponent(IStateMachine componentMainBehavior) {
            components.remove(componentMainBehavior.getId());
        }

        public void registerSession(IStateMachine outerComponent, String partName, IStateMachine session,
                String sessionId) {
            ComponentEntry c = getComponent(outerComponent.getId());
            if (c == null) {
                debug("Component with ID '" + outerComponent.getId() + "' is not registered yet!", "registerSession");
                return;
            }

            c.addSession(partName, session, sessionId);
            outerComponents.put(session.getId(), outerComponent.getId());
            partNames.put(session.getId(), partName);
            sessionIds.put(session.getId(), sessionId);
            sessions.put(sessionId, session);
        }

        public void removeSession(IStateMachine session) {
            String outerComponentId = getOuterComponentId(session);
            ComponentEntry c = getComponent(outerComponentId);
            if (c == null) return;
            String partName = getPartName(session);

            c.removeSession(partName, session);
            outerComponents.remove(session.getId());
            partNames.remove(session.getId());
            String sessionId = (String)sessionIds.remove(session.getId());
            sessions.remove(sessionId);
        }

        public Vector getSessions(IStateMachine outerComponent, String partName) {
            ComponentEntry c = getComponent(outerComponent.getId());
            if (c == null) return null;
            return c.getSessions(partName);
        }

        public IStateMachine getOuterComponent(IStateMachine session) {
            String outerComponentId = (String)outerComponents.get(session.getId());
            if (outerComponentId == null) return null;
            ComponentEntry c = getComponent(outerComponentId);
            return c.main;
        }

        public String getSessionIdOfStateMachine(IStateMachine session) {
            return (String)sessionIds.get(session.getId());
        }

        public IStateMachine getSession(IStateMachine outerComponent, String partName, String sessionId) {
            ComponentEntry ce = getComponent(outerComponent.getId());
            if (ce == null) return null;
            IStateMachine session = ce.getSession(partName, sessionId);
            if (session == null)
                return null;
            else return session;
        }

        private ComponentEntry getComponent(String componentId) {
            return (ComponentEntry)components.get(componentId);
        }

        private String getOuterComponentId(IStateMachine session) {
            return (String)outerComponents.get(session.getId());
        }

        private String getPartName(IStateMachine session) {
            return (String)partNames.get(session.getId());
        }

        protected void debug(String text, String method) {
            if (debug) System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
        }

        private static class ComponentEntry {

            final IStateMachine main;

            /*
             * Hashtable<String partName, PartEntry part>
             */
            private final Hashtable innerParts;

            public ComponentEntry(IStateMachine main) {
                this.main = main;
                innerParts = new Hashtable();
            }

            public IStateMachine getSession(String partName, String sessionId) {
                PartEntry pe = (PartEntry)innerParts.get(partName);
                if (pe == null) return null;
                for (int i = 0; i < pe.sessionIds.size(); i++) {
                    if (pe.sessionIds.elementAt(i).equals(sessionId)) {
                        return (IStateMachine)pe.sessions.elementAt(i);
                    }
                }
                return null;
            }

            public void addSession(String partName, IStateMachine session, String sessionId) {
                if (innerParts.get(partName) == null) innerParts.put(partName, new PartEntry());
                ((PartEntry)innerParts.get(partName)).addSession(session, sessionId);
            }

            public void removeSession(String partName, IStateMachine session) {
                PartEntry pe = (PartEntry)innerParts.get(partName);
                if (pe == null) return;
                // TODO: Could speed up by adding a hashtable instance
                pe.removeSession(session);
            }

            public Vector getSessions(String partName) {
                PartEntry p = (PartEntry)innerParts.get(partName);
                if (p != null) return p.sessions;
                else return null;
            }

            private static class PartEntry {

                final Vector sessions;
                final Vector sessionIds;

                public PartEntry() {
                    sessions = new Vector();
                    sessionIds = new Vector();
                }

                public void addSession(IStateMachine session, String sessionId) {
                    if (sessions.contains(session) || sessionIds.contains(sessionId)) {
                        if (debug)
                            System.out.println(this.getClass().getName() + ".addSession" + ":\n\t"
                                    + "Session or sessionId already added to this part, not added again.");
                        return;
                    }
                    sessions.addElement(session);
                    sessionIds.addElement(sessionId);
                }

                public void removeSession(IStateMachine session) {
                    int i = sessions.indexOf(session);
                    if (i != -1) {
                        sessions.removeElementAt(i);
                        sessionIds.removeElementAt(i);
                    }
                }
            }
        }
    }

    static private class Message {

        public final String signalID;
        public final Object sender;
        public final Object receiver;
        public final Object sessionId;
        public final Object content;
        public boolean marked;

        public Message(Object sender, Object receiver, Object sessionId, String signalID, Object content) {
            this.sender = sender;
            this.receiver = receiver;
            this.sessionId = sessionId;
            this.content = content;
            this.signalID = signalID;
            marked = false;
        }

        public String toString() {
            String n = System.getProperty("line.separator");
            StringBuffer s = new StringBuffer();
            s.append(n + "SignalID: " + signalID);
            s.append(n + "Sender: " + sender);
            s.append(n + "Receiver: " + receiver);
            s.append(n + "Session Id: " + sessionId);
            s.append(n + "Content :" + content);
            s.append(n + "Marked: " + marked);
            return s.toString();
        }

        public boolean isMarked() {
            return marked;
        }

        public void markAsSaved(boolean b) {
            marked = b;
        }
    }
}
