/**
 * 26. jan.. 2009
 */
package no.ntnu.item.ramses.runtime.transport.socket;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.nio.charset.Charset;

import no.ntnu.item.ramses.runtime.logging.Logger;
import no.ntnu.item.ramses.runtime.transport.Transporter;

/**
 * @author Marius Bjerke
 * @see Transporter
 */
public class SocketTransporter extends Transporter {

    /*
     * TODO: Currently relies on newLine between messages on the socket streams.
     * Check for (and remove?) newline in text, or introduce some special new
     * message char.. In the meantime, just pray that no message contains
     * newline chars!
     */

    /**
     * The {@link InetAddress} of this transporter.
     */
    private final InetAddress myInetAddress;

    /**
     * The port number used when listening for incoming connections.
     */
    private final int serverPort;

    private SocketCache socketCache;

    private ServerSocketWorker serverSocketWorker;

    /**
     * Constructor, sets {@link #myInetAddress}, does <b>not</b> call
     * {@link #initialize()}.
     */
    public SocketTransporter(int serverPort) {
        super();

        myInetAddress = getInetAddress();
        this.serverPort = serverPort;
    }

    public void initialize() {
        socketCache = new SocketCache();
        try {
            serverSocketWorker = new ServerSocketWorker();
            new Thread(serverSocketWorker).start();
        }
        catch (IOException e) {
            Logger.getLogger().communicationException(
                    "Failed to start listening for incoming connections!", e);

            debug("Failed to start listening for incoming connections!", "initialize");
            e.printStackTrace();
        }
    }

    public void closeTransporter() {
        serverSocketWorker.stopRunning();
        socketCache.emptyCacheAndCloseSockets();
    }

    public synchronized void transportMessage(String address, int port, String message) {
        /*
         * First check if we have an open connection to the destination.
         */
        if (socketCache.isCached(address, port)) {
            writeToSocket(address, port, message);
        }
        else {
            /*
             * We do not have an open connection, buffer the message, initiate
             * opening of a connection and the message should be sent when the
             * connection is opened.
             */
            unsentMessages.storeMessage(address, port, message);
            openConnection(address, port);
        }
    }

    public String getMyIpAddress() {
        return myInetAddress.getHostAddress();
    }

    public int getServerPort() {
        return serverPort;
    }

    protected void openConnection(String address, int port) {
        Socket newSocket = null;
        int numErr = 0;
        while (numErr < 3) {
            try {
                newSocket = new Socket(address, port);
                break;
            }
            catch (IOException e) {
                numErr++;
                Logger.getLogger().communicationException(
                        "Attempt failed to open Socket to " + address + ":" + port + ".", e);

                debug("Attempt failed to open Socket to " + address + ":" + port + ".",
                        "openConnection");

                if (newSocket != null && !newSocket.isClosed()) {
                    try {
                        newSocket.close();
                    }
                    catch (IOException e1) {
                        Logger.getLogger().communicationException(
                                "Tried to close not-closed socket after failed to open, even this threw exception! Ignoring.",
                                e);

                        debug(
                                "Tried to close not-closed socket after failed to open, even this threw exception! Ignoring..",
                                "openConnection");
                    }
                }

                if (numErr < 3) debug("Trying again..", "openConnection");
                e.printStackTrace();
            }
        }
        if (newSocket != null)
            connectionOpened(address, port, newSocket);
        else debug("Failed to open Socket to " + address + ":" + port
                + ", message might be left unsent in the buffer!", "openConnection");
    }

    protected void connectionOpened(String address, int port, Object connection) {
        Socket socket = (Socket)connection;

        // Store socket.
        SocketListenWorker listener = getAndStartSocketListenWorker(socket);
        socketCache.storeConnection(address, port, socket, listener);

        // Send unsent messages.
        String message;
        while (unsentMessages.hasBufferedMessageFor(address, port)) {
            message = unsentMessages.getAndRemoveOneMessage(address, port);
            writeToSocket(address, port, message);
        }
    }

    private InetAddress getInetAddress() {
        InetAddress addr = null;

        try {
            addr = InetAddress.getLocalHost();
        }
        catch (UnknownHostException e) {
            Logger.getLogger().communicationException("Could not find a local IP address.", e);
            debug("Could not find a local IP address!", "getInetAddress");
            e.printStackTrace();
        }

        return addr;
    }

    private void newIncomingSocket(Socket newSocket) {
        SocketListenWorker listener = getAndStartSocketListenWorker(newSocket);
        socketCache.storeConnection(SocketCache.getAddressKey(newSocket),
                SocketCache.getPortKey(newSocket), newSocket, listener);
    }

    private void writeToSocket(String address, int port, String message) {
        /*
         * TODO: Not use newLine as message separator, and then change from
         * PrintWriter to (Buffered)Writer, also in the SocketCache.
         */
        Writer w = socketCache.getConnectionWriter(address, port);

        try {
            w.write(message);
            w.write("\n");
            debug("Data written on a line to writer. Flushing..", "writeToSocket");
            w.flush();
            debug("Writer flushed.", "writeToSocket");
        }
        catch (IOException e) {
            Logger.getLogger().communicationException(
                    "Caught IOException while writing data to the writer, or flushing it.", e);
            debug("Caught IOException while writing data to the writer, or flushing it.",
                    "writeToSocket");
            // e.printStackTrace();
        }
        catch (Throwable e) {
            Logger.getLogger().communicationException(
                    "Caught throwable while writing data to the writer, or flushing it.", e);
        }
    }

    private SocketListenWorker getAndStartSocketListenWorker(Socket socket) {
        SocketListenWorker listener = null;
        try {
            listener = new SocketListenWorker(socket);
            new Thread(listener).start();
        }
        catch (IOException e) {
            Logger.getLogger().communicationException("IOException when creating socket listener",
                    e);
            debug(
                    "Caught IOException when creating SocketListenWorker, see SocketListenWorker(). State of socket now unknown!",
                    "getAndStartSocketListenWorker");
            // e.printStackTrace();
        }
        catch (Throwable e) {
            Logger.getLogger().communicationException("IOException when creating socket listener",
                    e);
        }
        return listener;
    }

    /**
     * A runnable class that listens continuously for incoming connections and
     * reports them back to {@link SocketTransporter#newIncomingSocket(Socket)}.
     * 
     * @author Marius Bjerke
     */
    private class ServerSocketWorker implements Runnable {

        private final ServerSocket serverSocket;
        private boolean isRunning;

        public ServerSocketWorker() throws IOException {
            serverSocket = new ServerSocket(serverPort);
            serverSocket.setSoTimeout(3000);
            isRunning = false;
        }

        public void run() {
            isRunning = true;
            while (isRunning) {
                try {
                    Socket incomingSocket = serverSocket.accept();
                    newIncomingSocket(incomingSocket);
                }
                catch (SocketTimeoutException e) {
                    // This is perfectly normal, and should happen every 3
                    // seconds, therefore no logging.
                    // Do nothing, continue looping.
                    // Logger.getLogger().communicationException("The server socket worker timed out.",
                    // e);
                }
                catch (IOException e) {
                	if (e instanceof SocketException && e.getMessage().equalsIgnoreCase("socket closed")) {
                		debug("ServerSocket closing..", "run");
                	}
                	else {
	                    Logger.getLogger().communicationException(
	                            "IOException caught while listening for incoming connections, listening continues.",
	                            e);
	                    debug(
	                            "IOException caught while listening for incoming connections, listening continues.",
	                            "run");
                	}
                }
            }
            debug("Stopped running.", "run");
        }

        public void stopRunning() {
            isRunning = false;
            try {
                serverSocket.close();
            }
            catch (IOException e) {
                Logger.getLogger().communicationException(
                        "IOException caught while closing server socket, ignoring.", e);
                debug("IOException caught while closing server socket, ignoring.", "stopRunning");
                // e.printStackTrace();
            }
        }
    }

    private class SocketListenWorker implements ConnectionListenWorker {

        private boolean isRunning;
        private Socket socket;
        private final BufferedReader socketReader;

        /**
         * Creates a SocketListenWorker.
         * 
         * @param socket
         *            The socket to listen on, must be connected.
         * @throws IOException
         *             If an I/O error occurs while getting input stream from
         *             socket.
         */
        public SocketListenWorker(Socket socket) throws IOException {
            if (!socket.isConnected())
                throw new IllegalStateException("Socket in SocketListenWorker must be connected!");

            this.socket = socket;
            InputStreamReader inReader = new InputStreamReader(socket.getInputStream(),
                    Charset.forName("UTF-8"));
            socketReader = new BufferedReader(inReader);
            isRunning = false;
        }

        public void run() {
            isRunning = true;
            int cErr = 0;
            while (isRunning) {
                /*
                 * TODO: Not use newLine as message separator.
                 */
                try {
                    String s = socketReader.readLine();
                    if (s == null) {
                        socketCache.removeAndCloseConnectionEntry(
                                SocketCache.getAddressKey(socket), SocketCache.getPortKey(socket));
                    }
                    else {
                        cErr = 0;
                        deliverMessageToRouter(s);
                    }
                }
                catch (IOException e) {
                	if (e.getMessage().equalsIgnoreCase("socket closed")) {
                		debug("Socket was closed.", "run");
                	}
                	else if (e.getMessage().equalsIgnoreCase("Connection reset")) {
                        debug("Connection was reset at another communication instance, closing its connection.", "run");
                        socketCache.removeAndCloseConnectionEntry(SocketCache.getAddressKey(socket), SocketCache.getPortKey(socket));
                    }
                    else {
                        Logger.getLogger().communicationException("IOException in socket listener", e);
                        cErr++;
                        debug("Caught " + cErr
                                + " consecutive IOException while trying readLine from socket.",
                                "SocketListenWorker.run");
    
                        if (cErr > 2) {
                            debug(
                                    "3 or more consecutive errors, closing the socket and removing it from the cache.",
                                    "run");
                            socketCache.removeAndCloseConnectionEntry(
                                    SocketCache.getAddressKey(socket), SocketCache.getPortKey(socket));
                        }
                        else debug("Trying again.", "run");
                    }
                }
            }
            debug("Stopped running.", "run");
        }

        /**
         * Stops the {@link SocketListenWorker} from trying to read another line
         * from the {@link #socketReader}.<br/>
         * <br/>
         * Does <b>not</b> close the socket.
         */
        public void stopRunning() {
            isRunning = false;
        }

    }

    /**
     * The SocketCache keeps track of sockets and relevant data, see
     * {@link ConnectionCache}.
     * 
     * @author Marius Bjerke
     */
    private static class SocketCache extends ConnectionCache {

        /**
         * @param socket
         *            The socket to get the address part of the key from.
         * @return The address part of the key to the {@link Socket}.
         */
        public static String getAddressKey(Socket socket) {
            return socket.getInetAddress().getHostAddress();
        }

        /**
         * @param socket
         *            The socket to get the port part of the key from.
         * @return The port part of the key to the {@link Socket}.
         */
        public static int getPortKey(Socket socket) {
            return socket.getPort();
        }

        protected void closeConnection(Object connection) {
            Socket socket = (Socket)connection;
            try {
                socket.close();
            }
            catch (IOException e) {
                Logger.getLogger().communicationException(
                        "Caught IOException while closing socket, ignoring. Possible socket open?",
                        e);
                debug("Caught IOException while closing socket, ignoring. Possible socket open?",
                        "closeConnection");
                // e.printStackTrace();
            }
        }

        protected Writer getWriterForConnection(Object connection) {
            Socket socket = (Socket)connection;
            Writer w;

            OutputStreamWriter outWriter = null;
            try {
                outWriter = new OutputStreamWriter(socket.getOutputStream());
            }
            catch (IOException e) {
                Logger.getLogger().communicationException(
                        "Caught IOException while getting output stream of socket, possible nullpointer coming up!",
                        e);
                debug(
                        "Caught IOException while getting output stream of socket, possible nullpointer coming up!",
                        "getWriterForConnection");
                // e.printStackTrace();
            }
            w = new BufferedWriter(outWriter);
            return w;
        }

        /**
         * Type-safe
         * {@link Transporter.ConnectionCache#getConnection(String, int)}.
         * 
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @return The socket associated with address and port.
         */
        public Socket getSocket(String address, int port) {
            return (Socket)getConnection(address, port);
        }
    }
}
