/**
 * 16. jan.. 2009
 */
package no.ntnu.item.ramses.runtime.routing;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import no.ntnu.item.ramses.runtime.Scheduler;
import no.ntnu.item.ramses.runtime.logging.Logger;
import no.ntnu.item.ramses.runtime.serialization.Serializer;
import no.ntnu.item.ramses.runtime.serialization.Serializer.MessageObjectsWrapper;
import no.ntnu.item.ramses.runtime.transport.Transporter;

/**
 * @author Marius Bjerke
 */
public abstract class Router {

    protected final static boolean debug = true;

    protected final static String itemUrlPrefix = "item://";

    /**
     * A serializer that should be used to (de)serialize outgoing and incoming
     * messages.
     */
    protected final Serializer serializer;

    /**
     * A {@link LocalRoutingTable} keeping track of addresses and relevant
     * information known to this router.
     */
    protected final LocalRoutingTable localRoutingTable;

    /**
     * A value indicating if the router is initialized and ready to route
     * messages.
     */
    protected boolean routerIsInitialized = false;

    /**
     * <b>Must</b> be called by any inheriting routers (as it is the only
     * visible superclass constructor) to initialize a {@link Serializer} and
     * {@link LocalRoutingTable}.
     * 
     * @param platform
     *            The platform this router will be running on.
     */
    protected Router(Serializer serializer) {
        this.serializer = serializer;
        this.localRoutingTable = new LocalRoutingTable();
    }

    /**
     * Should perform the necessary operations to make the router ready for
     * operation, e.g. opening listening ports etc.<br/>
     * <br/>
     * <b>Make sure that {@link #routerIsInitialized} is set to true when the
     * router is initialized properly.</b>
     */
    protected abstract void initialize();

    /**
     * Registers a state machine and finds an available ID for it, ensuring no
     * state machines are given equal addresses and allowing the router to
     * forward any messages addressed to that state machine to the scheduler it
     * is running on.<br/>
     * <br/>
     * If the router is not initialized, this method will call
     * {@link #initialize()}.
     * 
     * @param stateMachineId
     *            The ID of the state machine that is being registered.
     * @param componentType
     *            The component type of the state machine that is being
     *            registered.
     * @param scheduler
     *            The scheduler that the state machine being registered is
     *            running on.
     * @return The ITEM URL to the state machine that was registered.
     */
    public URL[] registerMainStateMachine(String componentType, Scheduler scheduler) {
        if (!routerIsInitialized) initialize();

        int i = 0;
        URL url;
        do {
            url = getItemUrlForMainStateMachine(componentType + i++);
        }
        while (localRoutingTable.getSchedulerForUrl(url) != null);

        localRoutingTable.addEntry(url, scheduler);

        URL[] alternateUrls = addAlternateUrls(url, scheduler);
        if (alternateUrls == null) alternateUrls = new URL[0];
        URL[] urls = new URL[alternateUrls.length + 1];
        urls[0] = url;
        for (int j = 0; j < alternateUrls.length; j++) {
            urls[j + 1] = alternateUrls[j];
        }
        
        return urls;
    }

    public URL[] registerSessionStateMachine(String outerComponentSmId, String partName, String sessionId,
            Scheduler scheduler) {
        if (!routerIsInitialized) initialize();

        URL url = getItemUrlForSessionStateMachine(outerComponentSmId, partName, sessionId);
        localRoutingTable.addEntry(url, scheduler);
        
        URL[] alternateUrls = addAlternateUrls(url, scheduler);
        if (alternateUrls == null) alternateUrls = new URL[0];
        URL[] urls = new URL[alternateUrls.length + 1];
        urls[0] = url;
        for (int j = 0; j < alternateUrls.length; j++) {
            urls[j + 1] = alternateUrls[j];
        }

        return urls;
    }

    /**
     * Must add all possible URLs pointing to the same component to the local
     * routing table. E.g. for IP addresses, it must include all the IP
     * addresses that points to the same endpont (129.241.. and 127.0.0.1,
     * etc.).<br/>
     * <br/>
     * If no alternate URLs exist, this method can be empty.<br/>
     * <br/>
     * The extract-from-address methods, (e.g.
     * {@link #extractComponentId(Object)}) can be used to find the information
     * in the passed url.
     * 
     * @param url
     */
    protected abstract URL[] addAlternateUrls(URL url, Scheduler scheduler);

    public void deregisterStateMachine(Object stateMachineUrl) {
        if (!(stateMachineUrl instanceof URL))
            throw new IllegalArgumentException("stateMachineUrl must be of type URL.");
        URL url = (URL)stateMachineUrl;

        localRoutingTable.removeEntry(url);
    }

    /**
     * Should return true if the url is addressed to this router, depending on
     * the ip address, port and/or similar.
     * 
     * @param url
     *            The {@link URL} to analyse
     * @return true if the address points to this router, false otherwise.
     */
    protected abstract boolean isThisRouter(URL url);

    /**
     * The method used to route a message. It will first check whether it can
     * forward the message to an appropriate {@link Scheduler}, if not it will
     * send the message to
     * {@link #routeOutgoingExternalMessage(URL, URL, String, Object)}.
     * 
     * @param sender
     *            The sender address, must be of type {@link URL}.
     * @param receiver
     *            The receiver address, must be of type {@link URL} or a
     *            {@link String} that can be parsed to one.
     * @param signalId
     *            The signal ID.
     * @param data
     *            The data object contained in the message.
     */
    public final void routeMessage(Object senderAddress, Object componentAddress, String partName, String sessionId,
            boolean allowCreateSession, String signalId, Object data) {
        debug("Entering..", "routeMessage");

        if (!(senderAddress instanceof URL && componentAddress instanceof URL))
            throw new IllegalArgumentException("Objects senderAddress and componentAddress must be of type URL.");
        URL senderUrl = (URL)senderAddress;
        URL componentUrl = (URL)componentAddress;

        debug("Routing: " + signalId + ", data=" + (data != null), "routeMessage");
        debug("Routing input: " + componentUrl + ", " + partName + ", " + sessionId + ", " + allowCreateSession + ", ",
                "routeMessage");

        if (!isThisRouter(componentUrl)) {
            debug("Not this router, routing out..", "routeMessage");
            URL receiverUrl = getCompleteUrl(componentUrl, partName, sessionId, allowCreateSession);
            routeOutgoingExternalMessage(senderUrl, receiverUrl, signalId, data);
        }
        else {
            Scheduler scheduler = localRoutingTable.getSchedulerForUrl(componentUrl);
            if (scheduler == null) {
                debug("Could not find scheduler for component, msg dropped.", "routeMessage");
                Logger.getLogger().messageDroppedDuringRouting(senderUrl.toString(), componentUrl.toString(), signalId);
            }
            else {
                scheduler.deliver(senderUrl, componentUrl, partName, sessionId, allowCreateSession, signalId, data,
                        true);
            }
        }

        debug("Leaving..", "routeMessage");
    }

    /**
     * Routes a message that is addressed to a state machine not registered in
     * this router. The method must evaluate the {@link URL}s, serialize the
     * message and forward the message to an appropriate {@link Transporter} .<br/>
     * <br/>
     * If the router is unable to forward the message appropriately, the message
     * is dropped. Note that if the appropriate next router for this message is
     * <b>this</b> router, it should also be dropped since the message was
     * passed from {@link #routeMessage(Object, Object, String, Object)} where
     * the local receiver could not be found.
     * 
     * @param sender
     *            The sender address.
     * @param receiver
     *            The receiver address.
     * @param signalId
     *            The signal ID.
     * @param data
     *            The data object contained in the message.
     * @see #serializer
     */
    protected abstract void routeOutgoingExternalMessage(URL sender, URL receiver, String signalId, Object data);

    /**
     * Receives an incoming external message. The incoming message is serialized
     * as a {@link String}, and must be deserialized before it can be delivered
     * to {@link #routeMessage(Object, Object, String, Object)}.
     * 
     * @param message
     *            The serialized incoming message.
     */
    public void routeIncomingExternalMessage(String serializedMessage) {
        debug("Incoming message: " + serializedMessage, "routeIncomingExternalMessage");

        MessageObjectsWrapper mow = serializer.deserialize(serializedMessage);

        if (mow == null) {
            debug("Deserialization failed, message dropped.", "routeIncomingExternalMessage");
            return;
        }

        Object senderAddress = mow.getSender();
        Object componentAddress = extractComponentAddress(mow.getReceiver());
        String partName = extractPartName(mow.getReceiver());
        String sessionId = extractSessionId(mow.getReceiver());
        boolean allowCreateSession = extractAllowCreateSession(mow.getReceiver());
        String signalId = mow.getSignalId();
        Object data = mow.getData();

        routeMessage(senderAddress, componentAddress, partName, sessionId, allowCreateSession, signalId, data);

        debug("Leaving..", "routeIncomingExternalMessage");
    }

    /**
     * Creates an address object from a string.
     * 
     * @param urlString
     *            The URL address as a String.
     * @return An address object.
     */
    public URL createAddress(String urlString) {
        return new URL(urlString);
    }

    public URL getCompleteUrl(Object componentAddress, String partName, String sessionId, boolean allowCreateSession) {
        if (!(componentAddress instanceof URL))
            throw new IllegalArgumentException("componentAddress must of type URL.");

        URL componentUrl = (URL)componentAddress;
        StringBuilder urlBuilder = new StringBuilder();
        urlBuilder.append(componentUrl.toString());
        if (partName != null && !partName.equals("")) {
            urlBuilder.append('/');
            urlBuilder.append(partName);
        }
        if (sessionId != null) {
            urlBuilder.append('?');
            urlBuilder.append(GlobalRoutingConstants.URL_QUERY_KEY_SESSION_ID);
            urlBuilder.append('=');
            urlBuilder.append(sessionId);
            if (allowCreateSession) {
                urlBuilder.append('&');
                urlBuilder.append(GlobalRoutingConstants.URL_QUERY_ALLOW_NEW);
                urlBuilder.append('=');
                urlBuilder.append("true");
            }
        }

        return new URL(urlBuilder.toString());
    }

    /**
     * Gets the ITEM URL to a main state machine with the corresponding state
     * machine ID.<br/>
     * <br/>
     * For implementations: This must be a complete ITEM address valid for the
     * state machine at its endpoint.<br/>
     * <br/>
     * This might be platform-specific.
     * 
     * @param mainStateMachineId
     *            The main state machine ID used for the {@link URL}.
     * @return The ITEM URL that points to the given main state machine at its
     *         endpoint.
     */
    protected abstract URL getItemUrlForMainStateMachine(String mainStateMachineId);

    protected URL getItemUrlForSessionStateMachine(String outerComponentSmId, String partName, String sessionId) {
        URL outerComponentUrl = getItemUrlForMainStateMachine(outerComponentSmId);
        return getCompleteUrl(outerComponentUrl, partName, sessionId, false);
    }

    /**
     * Notifies the router that scheduler is stopped, i.e. it removes all
     * entries in the routing table that are associated with scheduler. It also
     * calls {@link #notifySchedulerStoppedPlatformSpecific(Scheduler)} allowing
     * the router to close any platform specific resources if there are no more
     * schedulers running.
     * 
     * @param scheduler
     *            The {@link Scheduler} that has been stopped.
     */
    public void notifySchedulerStopped(Scheduler scheduler) {
        localRoutingTable.removeEntries(scheduler);

        notifySchedulerStoppedPlatformSpecific(scheduler);
    }

    /**
     * If all {@link Scheduler}s known to be using this router is stopped, make
     * sure that all external resources like sockets etc. are closed.
     * 
     * @param scheduler
     */
    protected abstract void notifySchedulerStoppedPlatformSpecific(Scheduler scheduler);

    public abstract URL extractComponentAddress(Object address);

    public String extractComponentId(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;

        URL componentAddress = extractComponentAddress(url);
        return (String)componentAddress.getPathComponents().lastElement();
    }

    public String extractSessionSmId(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;

        return extractComponentId(url) + "/" + extractPartName(url) + "?"
                + GlobalRoutingConstants.URL_QUERY_KEY_SESSION_ID + "=" + extractSessionId(url);
    }

    public abstract String extractPartName(Object address);

    public String extractSessionId(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;

        if (url.getQueryMap() != null) {
            return (String)url.getQueryMap().get(GlobalRoutingConstants.URL_QUERY_KEY_SESSION_ID);
        }
        else {
            return null;
        }
    }

    public boolean extractAllowCreateSession(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;

        if (url.getQueryMap() != null) {
            Object allowNew = url.getQueryMap().get(GlobalRoutingConstants.URL_QUERY_ALLOW_NEW);
            if (allowNew != null) return allowNew.equals("true");
        }
        return false;
    }

    protected void debug(String text, String method) {
        if (debug) System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
    }

    /**
     * The LocalRoutingTable keeps track of which state machine instances are
     * located at what {@link Scheduler}s. It also keeps track of state machine
     * instance by their component type.
     * 
     * @author Marius Bjerke
     */
    protected class LocalRoutingTable {

        /**
         * Maps URLs to a Scheduler.<br/>
         * Hashtable<URL, Scheduler>
         */
        private final Hashtable urlToScheduler;

        /**
         * Keeps a map of active schedulers and {@link URL}s to their active
         * state machines.<br/>
         * Hashtable<Scheduler, Vector<URL>>
         */
        private final Hashtable schedulerToUrls;

        public LocalRoutingTable() {
            urlToScheduler = new Hashtable();
            schedulerToUrls = new Hashtable();
        }

        /**
         * @return true if the LocalRoutingTable is empty, false otherwise.
         */
        public boolean isEmpty() {
            return urlToScheduler.isEmpty();
        }

        /**
         * Adds an entry for stateMachineUrl to the LocalRoutingTable.
         * 
         * @param stateMachineUrl
         *            The {@link URL} key.
         * @param scheduler
         *            The associated {@link Scheduler}.
         */
        public void addEntry(URL stateMachineUrl, Scheduler scheduler) {
            urlToScheduler.put(stateMachineUrl, scheduler);

            Vector v = (Vector)schedulerToUrls.get(scheduler);
            if (v == null) {
                v = new Vector();
                schedulerToUrls.put(scheduler, v);
            }
            v.addElement(stateMachineUrl);
        }

        /**
         * Removes the mapping for stateMachineUrl from the LocalRoutingTable,
         * does nothing if this mapping did not exist.
         * 
         * @param stateMachineUrl
         *            The {@link URL} to remove from the table.
         */
        public void removeEntry(URL stateMachineUrl) {
            Scheduler scheduler = (Scheduler)urlToScheduler.remove(stateMachineUrl);

            if (scheduler != null) {
                Vector v = (Vector)schedulerToUrls.get(scheduler);
                if (v != null) v.removeElement(stateMachineUrl);
                if (v.isEmpty()) schedulerToUrls.remove(scheduler);
            }
        }

        /**
         * Removes all the {@link URL}s to state machines running on the given
         * scheduler from the LocalRoutingTable.
         * 
         * @param scheduler
         *            The {@link Scheduler} for which all associated {@link URL}
         *            s should be removed.
         */
        public void removeEntries(Scheduler scheduler) {
            Vector v = (Vector)schedulerToUrls.get(scheduler);
            if (v != null) {
                while (!v.isEmpty()) {
                    removeEntry((URL)v.firstElement());
                }
            }
        }

        /**
         * Returns the {@link Scheduler} associated with stateMachineUrl, or
         * null if no such association exists.
         * 
         * @param stateMachineUrl
         *            The {@link URL} to find the Scheduler for.
         * @return A Scheduler associated with stateMachineUrl, or null if none
         *         exists.
         */
        public Scheduler getSchedulerForUrl(URL stateMachineUrl) {
            return (Scheduler)urlToScheduler.get(stateMachineUrl);
        }
    }
}
