/**
 * 16. jan.. 2009
 */
package no.ntnu.item.ramses.runtime.routing.javase;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;

import no.ntnu.item.ramses.runtime.Scheduler;
import no.ntnu.item.ramses.runtime.logging.Logger;
import no.ntnu.item.ramses.runtime.routing.GlobalRoutingConstants;
import no.ntnu.item.ramses.runtime.routing.Router;
import no.ntnu.item.ramses.runtime.routing.URL;
import no.ntnu.item.ramses.runtime.serialization.Serializer;
import no.ntnu.item.ramses.runtime.transport.socket.SocketTransporter;

/**
 * The JavaStandardRouter is meant to run a regular computer, with TCP/IP
 * capabilities.<br/>
 * <br/>
 * It will be addressable by it's IP address, and listen for incoming
 * connections on port
 * {@link GlobalRoutingConstants#JAVA_STANDARD_TCP_INCOMING_PORT}.
 * 
 * @author Marius Bjerke
 */
public class JavaStandardRouter extends Router {

    /**
     * The {@link SocketTransporter} used by this router to transport messages.
     */
    private final SocketTransporter socketTransporter;

    /**
     * The IP address of this router.
     */
    private String routerIpAddress;

    /**
     * The port of this router.
     */
    private int incomingListenPort;

    public JavaStandardRouter(Serializer serializer, SocketTransporter transporter) {
        super(serializer);

        incomingListenPort = transporter.getServerPort();
        socketTransporter = transporter;
        routerIpAddress = socketTransporter.getMyIpAddress();

        transporter.setRouter(this);

        initialize();
    }

    /**
     * @see {@link Router#initialize()}
     */
    protected void initialize() {
        socketTransporter.initialize();

        routerIsInitialized = true;
    }

   /**
    * @see {@link Router#addAlternateUrls(URL)}
    */
    protected URL[] addAlternateUrls(URL url, Scheduler scheduler) {
        ArrayList<URL> urlList = new ArrayList<URL>();
        URL loopback = exchangeIpAddress(url, "127.0.0.1");
        if (localRoutingTable.getSchedulerForUrl(loopback) == null) {
            debug("Also adding to routing table: " + loopback, "addAlternateUrls");
            localRoutingTable.addEntry(loopback, scheduler);
            urlList.add(loopback);
        }
        try {
            InetAddress localhost = InetAddress.getLocalHost();
            // Just in case this host has multiple IP addresses....
            InetAddress[] allMyIps = InetAddress.getAllByName(localhost.getCanonicalHostName());
            if (allMyIps != null) {
                for (int i = 0; i < allMyIps.length; i++) {
                    if (allMyIps[i] instanceof Inet4Address) {
                        URL newUrl = exchangeIpAddress(url, allMyIps[i].getHostAddress());
                        if (localRoutingTable.getSchedulerForUrl(newUrl) == null) {
                            debug("Also adding to routing table: " + newUrl, "addAlternateUrls");
                            localRoutingTable.addEntry(newUrl, scheduler);
                            urlList.add(newUrl);
                        }
                    }
                }
            }
        } catch (UnknownHostException e) {
            // Ignore..
        }
        return urlList.toArray(new URL[0]);
    }
    
    private URL exchangeIpAddress(URL url, String ip) {
        StringBuilder sb = new StringBuilder();
        sb.append(url.getProtocol());
        sb.append("://");
        sb.append(ip);
        if (!url.getPort().equals("")) {
            sb.append(':');
            sb.append(url.getPort());
        }
        for (Object p : url.getPathComponents()) {
            String s = (String)p;
            sb.append('/');
            sb.append(s);
        }
        if (!url.getQuery().equals("")) {
            sb.append('?');
            sb.append(url.getQuery());
        }
        return new URL(sb.toString());
    }

    /**
     * @see {@link Router#isThisRouter(URL)}
     */
    protected boolean isThisRouter(URL url) {
        if (hasThisIpAddress(routerIpAddress) && portMatches(url)) return true;
        else return false;
    }
    
    private boolean hasThisIpAddress(String ip) {
        try {
            InetAddress.getByName(ip);
        }
        catch (UnknownHostException e) {
            return false;
        }
        return true;
    }

    protected boolean portMatches(URL url) {
        if (incomingListenPort == GlobalRoutingConstants.JAVA_STANDARD_TCP_SERVER_PORT) {
            if (("").equals(url.getPort()) || ("" + incomingListenPort).equals(url.getPort())) return true;
        }
        else if (("" + incomingListenPort).equals(url.getPort())) return true;
        return false;
    }

    /**
     * @see {@link Router#routeOutgoingExternalMessage(URL, URL, String, Object)}
     */
    protected void routeOutgoingExternalMessage(URL sender, URL receiver, String signalId,
            Object data) {
//        /*
//         * First check if the next router is this router, if it is the message
//         * should be dropped.
//         */
//        if (isThisRouter(receiver)) {
//            /*
//             * This router is the next hop, drop the message.
//             */
//            Logger.getLogger().messageDroppedDuringRouting(sender.toString(), receiver.toString(), signalId);
//        }
//        else {
            /*
             * Then validate (simply) that the URL seems to have a valid IP address,
             * and if so, it can be serialized and delivered to the transporter.
             */
            if (receiver.getAddress().length() > "x.x.x.x".length()) {
                String serializedMessage = serializer.serialize(sender, receiver, signalId, data);
                int port = GlobalRoutingConstants.JAVA_STANDARD_TCP_SERVER_PORT; 
                if (!receiver.getPort().equals("")) port = Integer.parseInt(receiver.getPort());
                socketTransporter.transportMessage(receiver.getAddress(), port,
                        serializedMessage);
            }
            else {
                debug("Recipient address did not pass minimum length criteria. Dropping message!",
                        "routeOutgoingExternalMessage");
                Logger.getLogger().messageDroppedDuringRouting(sender.toString(), receiver.toString(), signalId);
            }
//        }
    }

    /**
     * @see {@link Router#getItemUrlForMainStateMachine(String)}
     */
    protected URL getItemUrlForMainStateMachine(String stateMachineId) {
        StringBuffer sb = new StringBuffer();

        sb.append("item://");
        sb.append(routerIpAddress);
        if (incomingListenPort != GlobalRoutingConstants.JAVA_STANDARD_TCP_SERVER_PORT) {
            sb.append(":");
            sb.append(String.valueOf(incomingListenPort));
        }
        sb.append("/");
        sb.append(stateMachineId);

        return new URL(sb.toString());
    }

    /**
     * see {@link Router#notifySchedulerStopped(Scheduler)}
     */
    protected void notifySchedulerStoppedPlatformSpecific(Scheduler scheduler) {
        if (localRoutingTable.isEmpty()) {
            /*
             * Let's close down shop.
             */
            socketTransporter.closeTransporter();
        }
    }

    /**
     * @see {@link Router#extractComponentAddress(Object)}
     */
    public URL extractComponentAddress(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;
        
        if (url.getPathComponents().size() == 1 && url.getQuery().equals("") && url.getFragment().equals("")) {
            // There is no other parts than the state machine ID in the path or later.
            return url;
        }
        
        StringBuilder componentUrl = new StringBuilder();
        componentUrl.append(url.getProtocol());
        componentUrl.append("://");
        componentUrl.append(url.getAddress());
        if (!url.getPort().equals("")) {
            componentUrl.append(':');
            componentUrl.append(url.getPort());
        }
        componentUrl.append('/');
        componentUrl.append(url.getPathComponents().firstElement());
        
        return new URL(componentUrl.toString());
    }

    /**
     * @see {@link Router#extractPartName(Object)}
     */
    public String extractPartName(Object address) {
        if (!(address instanceof URL)) throw new IllegalArgumentException("Address must be an URL object.");
        URL url = (URL)address;
        
        if (url.getPathComponents().size() == 2) return (String)url.getPathComponents().lastElement();
        else return null;
    }
}
