/**
 * 26. jan.. 2009
 */
package no.ntnu.item.ramses.runtime.transport;

import java.io.IOException;
import java.io.Writer;
import java.util.Hashtable;
import java.util.Vector;

import no.ntnu.item.ramses.runtime.routing.Router;

/**
 * This abstract Transporter superclass provides some very basic framework for
 * subclassing Transporters, as well as specifying some methods that every
 * Transporter implementation must have. Specifically these methods are:<br/>
 * <ol>
 * <li>initialize() - See {@link #initialize()}.
 * <li>transportMessage(String address, int port, String message) - See
 * {@link #transportMessage(String, int, String)}.
 * </ol>
 * 
 * @author Marius Bjerke
 */
public abstract class Transporter {

    protected static final boolean debug = true;

    /**
     * The {@link Router} that owns this Transporter. Messages received will be
     * forwarded to this router. Should this one be protected, or is there no
     * need?
     */
    private Router router;

    /**
     * An UnsentMessageBuffer for messages that are awaiting an open connection.
     * The pair of address and port is the key to the buffer, and points to a
     * list containing messages to that destination.
     * 
     * @see UnsentMessageBuffer
     */
    protected final UnsentMessageBuffer unsentMessages;

    /**
     * <b>Must</b> be called by any inheriting class, assigning the field
     * {@link #router}.
     * 
     * @param router
     *            The {@link Router} owning this Transporter.
     */
    protected Transporter() {
        unsentMessages = new UnsentMessageBuffer();
    }

    /**
     * Initializes the transporter. When this method returns, it should be ready
     * to transport messages, and able to receive them. Things every subclass of
     * {@link Transporter} should consider initializing includes:<br>
     * <ol>
     * <li>An implementation of {@link ConnectionCache}.
     * <li>An implementation of {@link ServerConnectionWorker}.
     * </ol>
     */
    public abstract void initialize();

    /**
     * Stops the Transporter object, releasing all tied up resources like open
     * sockets, etc. Things every subclass of {@link Transporter} should
     * consider closing includes:<br>
     * <ol>
     * <li>Entries in a {@link ConnectionCache}.
     * <li>A running {@link ServerConnectionWorker}.
     * </ol>
     */
    public abstract void closeTransporter();

    /**
     * This should be the method used by the router to send a message. The
     * meaning of the parameters depend on the implementation of Transporter,
     * e.g. a {@link XXXJavaStandardRouter} would need IP addresses, and some
     * transporters might not need a port at all.
     * 
     * @param address
     *            The address the message should be delivered to, as a
     *            {@link String} representation defined by the
     *            {@link Transporter}-implementation.
     * @param port
     *            The port the message should be delivered to, if this is
     *            required by the {@link Transporter}-implementation. If it is
     *            not required, the recommended value is -1.
     * @param message
     *            The message to be delivered, serialized as a {@link String}.
     */
    public abstract void transportMessage(String address, int port, String message);

    public void setRouter(Router router) {
        this.router = router;
    }

    protected abstract void openConnection(String address, int port);

    protected abstract void connectionOpened(String address, int port, Object connection);

    /**
     * Delivers a received serialized message to the owning router.
     * 
     * @param message
     *            The received serialized message.
     */
    protected void deliverMessageToRouter(String serializedMessage) {
        router.routeIncomingExternalMessage(serializedMessage);
    }

    protected void debug(String text, String method) {
        if (debug) System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
    }

    /**
     * An object for storing unsent messages to a destination decided by the
     * String address and int port. These two together form the key to the
     * buffer.<br/>
     * <br/>
     * If port is meaningless for a Transporter implementation, it could
     * (should?) use the constant {@link UnsentMessageBuffer#NO_PORT} for the
     * port.
     * 
     * @author Marius Bjerke
     */
    protected static class UnsentMessageBuffer {

        /**
         * An int that can be used by transporters where a port no. is
         * meaningless, to ensure consistency.
         */
        public static final int NO_PORT = 0;

        /**
         * The buffer.<br>
         * Hashtable<getBufferKey(String, int), Vector<String>>
         */
        private final Hashtable buffer;

        /**
         * Creates an empty buffer.
         */
        public UnsentMessageBuffer() {
            buffer = new Hashtable();
        }

        /**
         * @param address
         *            The address part of the buffer key.
         * @param port
         *            The port part of the buffer key.
         * @return true if a message is buffered for the given address/port,
         *         false otherwise.
         */
        public boolean hasBufferedMessageFor(String address, int port) {
            String bufferKey = getBufferKey(address, port);
            if (buffer.containsKey(bufferKey))
                return true;
            else return false;
        }

        /**
         * Stores a message in the buffer.
         * 
         * @param address
         *            The address part of the buffer key.
         * @param port
         *            The port part of the buffer key.
         * @param message
         *            The message to put in the buffer.
         */
        public void storeMessage(String address, int port, String message) {
            String bufferKey = getBufferKey(address, port);
            Vector messageList = (Vector)buffer.get(bufferKey);
            if (messageList == null) {
                messageList = new Vector();
                buffer.put(bufferKey, messageList);
            }
            messageList.addElement(message);
        }

        /**
         * Removes and returns a message to the given address/port from the
         * buffer.
         * 
         * @param address
         *            The address part of the buffer key.
         * @param port
         *            The port part of the buffer key.
         * @return A message given by address and port, or null if no such
         *         message exists.
         */
        public String getAndRemoveOneMessage(String address, int port) {
            if (!hasBufferedMessageFor(address, port)) return null;

            Vector v = (Vector)buffer.get(getBufferKey(address, port));
            String message = (String)v.firstElement();
            v.removeElement(message);

            /*
             * Clean up vector if it is empty.
             */
            if (v.isEmpty()) buffer.remove(getBufferKey(address, port));

            return message;
        }

        private String getBufferKey(String address, int port) {
            return address + ":" + port;
        }
    }

    public interface ConnectionListenWorker extends Runnable {

        public void stopRunning();
    }

    /**
     * A class to keep track of open connections. Each Transporter should extend
     * this and add a type-safe method for returning the connection object, e.g.
     * for SocketTransporter it will be a Socket, and possibly a method for
     * getting a ConnectionCache key from a connection object.<br/>
     * <br/>
     * If port is meaningless for a Transporter implementation, it could
     * (should?) use the constant {@link UnsentMessageBuffer#NO_PORT} for the
     * port.
     * 
     * @author Marius Bjerke
     */
    protected static abstract class ConnectionCache {

        /**
         * An int that can be used by transporters where a port no. is
         * meaningless, to ensure consistency.
         */
        public static final int NO_PORT = 0;

        /**
         * A map to keep track of open connections and the associated
         * {@link SocketCacheEntry}<br/>
         * <br/>
         * Hashtable<getCacheKey(String, int), {@link SocketCacheEntry}>
         */
        private final Hashtable connectionCacheTable;

        /**
         * Initiates an empty ConnectionCache.
         */
        public ConnectionCache() {
            connectionCacheTable = new Hashtable();
        }

        protected static String getConnectionCacheKey(String address, int port) {
            return address + ":" + port;
        }

        /**
         * Stores a connection object and associates it with two part key
         * consisting of a String address and int port.
         * 
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @param connection
         *            The connection associated with the address and port, e.g.
         *            a Socket.
         */
        public synchronized void storeConnection(String address, int port, Object connection,
                ConnectionListenWorker listener) {
            ConnectionCacheEntry entry = new ConnectionCacheEntry(connection, listener);
            connectionCacheTable.put(getConnectionCacheKey(address, port), entry);
        }

        /**
         * Returns a connection object associated with the address and port, or
         * null if none exists.
         * 
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @return The connection object associated with the address and port.
         */
        public Object getConnection(String address, int port) {
            return ((ConnectionCacheEntry)connectionCacheTable.get(getConnectionCacheKey(address,
                    port))).getConnection();
        }

        /**
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @return The {@link ConnectionListenWorker} associated with address
         *         and port.
         */
        public ConnectionListenWorker getConnectionListenWorker(String address, int port) {
            return ((ConnectionCacheEntry)connectionCacheTable.get(getConnectionCacheKey(address,
                    port))).getListener();
        }

        /**
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @return A {@link Writer} for the connection associated with the
         *         address and port.
         */
        public Writer getConnectionWriter(String address, int port) {
            return ((ConnectionCacheEntry)connectionCacheTable.get(getConnectionCacheKey(address,
                    port))).getConnectionWriter();
        }

        /**
         * Removes, from the cache, and closes the {@link ConnectionCacheEntry}
         * associated with addressPort.
         * 
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         */
        public synchronized void removeAndCloseConnectionEntry(String address, int port) {
            ConnectionCacheEntry s = (ConnectionCacheEntry)connectionCacheTable.get(getConnectionCacheKey(
                    address, port));
            if (s != null) {
                connectionCacheTable.remove(getConnectionCacheKey(address, port));
                s.closeEntry();
            }
        }

        /**
         * @return true if the ConnectionCache contains no entries, false
         *         otherwise.
         */
        public boolean isEmpty() {
            return connectionCacheTable.isEmpty();
        }

        /**
         * @param address
         *            The address part of the key.
         * @param port
         *            The port part of the key.
         * @return true if an entry associated with address and port is in the
         *         cache.
         */
        public boolean isCached(String address, int port) {
            return connectionCacheTable.containsKey(getConnectionCacheKey(address, port));
        }

        /**
         * Empties the {@link ConnectionCache} and closes each entry, see
         * {@link ConnectionCacheEntry#closeEntry()}.
         */
        public synchronized void emptyCacheAndCloseSockets() {
            Object entryKey;
        	ConnectionCacheEntry entry;
            while (!this.isEmpty()) {
            	entryKey = connectionCacheTable.keys().nextElement();
                entry = (ConnectionCacheEntry)connectionCacheTable.get(entryKey);
                connectionCacheTable.remove(entryKey);
                entry.closeEntry();
            }
        }

        /**
         * This method should close the object connection.
         * 
         * @param connection
         *            The connection object to close, could be a Socket or
         *            similar.
         */
        protected abstract void closeConnection(Object connection);

        /**
         * This method should return a Writer for the output stream of
         * connection.
         * 
         * @param connection
         *            The object to get a Writer for.
         * @return A writer for the output stream of connection.
         */
        protected abstract Writer getWriterForConnection(Object connection);

        protected void debug(String text, String method) {
            if (debug)
                System.out.println(this.getClass().getName() + "." + method + ":\n\t" + text);
        }

        /**
         * The ConnectionCacheEntry contains the following:<br/>
         * <ol>
         * <li>The connection instance.
         * <li>A {@link ConnectionListenWorker} that is listening for text on
         * the socket.
         * <li>A {@link Writer} that is ready to write to the socket's output
         * stream.
         * </ol>
         * 
         * @author Marius Bjerke
         */
        private class ConnectionCacheEntry {

            private final Object connection;
            private final ConnectionListenWorker listener;
            private final Writer connectionWriter;

            public ConnectionCacheEntry(Object connection, ConnectionListenWorker listener) {
                this.connection = connection;
                this.listener = listener;
                this.connectionWriter = getWriterForConnection(connection);
            }

            public Object getConnection() {
                return connection;
            }

            public ConnectionListenWorker getListener() {
                return listener;
            }

            public Writer getConnectionWriter() {
                return connectionWriter;
            }

            /**
             * Closes all the resources in the entry, this includes:<br/>
             * <ol>
             * <li>Stopping the listener, see
             * {@link SocketListenWorker#stopRunning()}.
             * <li>Closing the
             */
            public synchronized void closeEntry() {
                listener.stopRunning();
                try {
                    connectionWriter.close();
                }
                catch (IOException e) {
                    debug("Caught IOException while closing Writer, ignoring.", "closeEntry");
                    // e.printStackTrace();
                }
                closeConnection(connection);
            }
        }
    }
}
